#!/usr/bin/env python
# coding=utf-8
"""
DICTOR testing suite
shell> pytest test.py
"""

from __future__ import print_function
import json
import sys
import pytest
from os.path import dirname, join, abspath

## import repo version of dictor, not pip-installed version
sys.path.insert(0, abspath(join(dirname(__file__), "..")))
from dictor import dictor

with open("basic.json") as data:
    BASIC = json.load(data)

with open("list.json") as data:
    LIST = json.load(data)

with open("large.json") as data:
    LARGE = json.load(data)

with open("nested.json") as data:
    NESTED = json.load(data)

def test_simple_list():
    """test for basic list value"""
    result = dictor(LIST, "1.name")
    assert result == "gone with the wind"

def test_simple_dict():
    """test for value in a dictionary"""
    result = dictor(BASIC, "robocop.year")
    assert result == 1989

def test_non_existent_value():
    """test a non existent key search"""
    result = dictor(BASIC, "non.existent.value")
    assert result is None

    result = dictor({"lastname": "Doe"}, "foo.lastname")
    assert result is None

def test_zero_value():
    """test a Zero value - should return 0"""
    result = dictor(BASIC, "terminator.2.terminator 3.year", checknone=True)
    assert result == 0

def test_partial_exist_value():
    """partially existing value"""
    result = dictor(BASIC, "spaceballs.year.fakekey")
    assert result is None

def test_random_chars():
    result = dictor(BASIC, "#.random,,,@.chars")
    assert result == '({%^&$"'

def test_complex_dict():
    """test parsing down a list and getting dict value"""
    result = dictor(BASIC, "terminator.1.terminator 2.genre.0")
    assert result == "nuclear war"

def test_pathsep():
    """test parsing down a list and getting dict value with pathsep"""
    result = dictor(BASIC, "terminator/1/terminator 2/genre/0", pathsep="/")
    assert result == "nuclear war"

def test_keys_with_different_pathsep():
    """test parsing keys with different path separator"""
    result = dictor(BASIC, "dirty.harry/genre", pathsep="/")
    assert result == "romance"

def test_escape_pathsep():
    """test using escape path separator"""
    result = dictor(BASIC, "dirty\.harry.genre")
    assert result == "romance"

def test_ignore_letter_casing():
    """test ignoring letter upper/lower case"""
    result = dictor(BASIC, "austin PoWeRs.year", ignorecase=True)
    assert result == 1996

def test_ignore_letter_casing_nested():
    """test ignoring letter upper/lower case"""
    result = dictor(BASIC, "austin PoWeRs.Year", ignorecase=True)
    assert result == 1996

def test_numeric_key_handling():
    """test handling keys that are numbers"""
    result = dictor(BASIC, "1492.year")
    assert result == 1986

def test_parsing_large_JSON():
    """test parsing large JSON file"""
    result = dictor(LARGE, "0.tags.3")
    assert result == "sunt"

    result = dictor(LARGE, "1.friends.2.name")
    assert result == "Tanisha Saunders"

def test_parsing_large_JSON_nonexistent():
    """test parsing large JSON file w nonexistent value"""
    result = dictor(LARGE, "some.value")
    assert result is None

def test_exception():
    """test for non existent index value"""
    with pytest.raises(ValueError):
        dictor(LIST, "5.genre", checknone=True)

def test_searching_list_JSON():
    """test searching list JSON file"""
    result = dictor(LIST, search="name")
    assert result == ["spaceballs", "gone with the wind", "titanic", "titanic"]

def test_searching_List_for_duplicate_key_value():
    """ test searching Json file for multiple identical key values """
    result = dictor(LIST, search="name")
    assert result == ["spaceballs", "gone with the wind", "titanic", "titanic"]

def test_searching_large_JSON():
    """test searching large JSON file"""
    result = dictor(LARGE, "0.friends", search="name")
    assert result == ["Patsy Sargent", "Bailey Carpenter", "Corina Sherman"]
    
    result = dictor(LARGE, search="company")
    assert result == ['VISUALIX', 'ZORK', 'GRAINSPOT', 'QUANTALIA', 'SUPREMIA', 'PURIA', 'BLEENDOT']

def test_return_type_int():
    """test returning int type"""
    result = dictor(BASIC, "conan the barbarian.year", rtype="int")
    assert result == 1983

def test_return_type_str():
    """test returning str type"""
    result = dictor(BASIC, "spaceballs.year", rtype="str")
    assert result == "1987"

def test_return_invalid_rtype():
    """test string to int invalid rtype"""
    result = dictor(BASIC, "spaceballs.genre", rtype="int")
    assert result == "comedy"

def test_empty_string_path():
    """test empty string path"""
    result = dictor(BASIC, "")
    assert result is None

def test_malformed_search_path():
    """test_malformed_search_path"""
    result = dictor(BASIC, "spaceballs.")
    assert result is None

def test_non_existent_path_search():
    """test_non_existent_path_search"""
    result = dictor(BASIC, "nonexistent.subkey", search="abc")
    assert result is None

    result = dictor(BASIC, "nonexistent.subkey", default="path doesnt exist", search="abc")
    assert result == "path doesnt exist"


def test_search_List_for_bools_with_default_fallback():
    """
    if json key is like this:

    { "key": false } --- this should return a python False boolean, not fallback default value

    if key has null, then should return a Default fallback value, and if not provided a fallback, return a None
    """
    result = dictor(LIST, search="status", default="fallback")
    assert result == [False, "", True, "fallback", "subkey1 status", "null"]


def test_search_List_for_bools_with_no_fallback():
    """
    if json key is like this:

    { "key": false } --- this should return a python False boolean, not fallback default value

    if key has value of null, then should return a None
    if key has value of "null", then should return "null" string
    """
    result = dictor(LIST, search="status")
    assert result == [False, "", True, None, "subkey1 status", "null"]

def test_search_Basic_for_bools_with_default_fallback():
    """
    if json key is like this:

    { "key": false } --- this should return a python False boolean, not fallback default value

    if key has null, then should return a Default fallback value, and if not provided a fallback, return a None
    """
    result = dictor(BASIC, search="status", default="fallback")
    assert result == [False, True, "fallback", "", "Nested Status"]

def test_search_for_word_null_with_default_fallback():
    """ testing a key named "null" """
    result = dictor(LARGE, search="null", default="fallback")
    assert result == ["fake_null", "fallback"]

def test_search_for_word_null_with_no_default_fallback():
    result = dictor(LARGE, search="null")
    assert result == ["fake_null", None]

def test_null_value():
    result = dictor(BASIC, "terminator.2.terminator 3.preview", default="Missing key")
    assert result is None


def test_empty_string_value():
    result = dictor(BASIC, "terminator.2.terminator 3.stars", default="Missing key")
    assert result == ""

def test_getting_tuple_value():
    data = {
        "name": "joe",
        "age": 32,
        "hobbies": ("skiing", "archery", "chess"),
        "foods": ["spam", "celery", ("milk", "cheese", "yogurt"), "cake"],
        "key1": {
            "foods": ["carrot", "potato"],
            "subkey1": {
                "foods": ("pear", "cherry")
            }
        }
    }
    result = dictor(data, "hobbies")
    assert result == ("skiing", "archery", "chess")
    
    # test getting tuple from index #2 of a list of elements
    result = dictor(data, "foods.2")
    assert result == ("milk", "cheese", "yogurt")

    # test searching for nested keys with tuple values
    result = dictor(data, search="foods")
    assert result == [['spam','celery', ('milk','cheese','yogurt'),'cake'], ['carrot','potato'], ('pear','cherry')]


def test_nested_dict_value():
    data = {
        "k1": {
            "k2": {
                "k3": {
                    "k4": {
                        "k5": {
                            "k6": {
                                "k7": {
                                    "k8": {
                                        "k9": {
                                            "k10": {},
                                            "level": "L10"
                                        }

                                    }
                                }
                            },
                            "level": "L6"
                        },
                        "level": "L5"
                    }
                }
            },
            "level": "L2"
        },
        "level": "L1"
    }
    result = dictor(data, search="level")
    assert result == ['L10', 'L6', 'L5', 'L2', 'L1']

    # empty dict declaration should return as dict, not as a None
    result = dictor(data, "k1.k2.k3.k4.k5.k6.k7.k8.k9.k10")
    assert isinstance(result, dict)

    # non existent key
    result = dictor(data, "k1.k2.k3.k4.k5.k6.k7.k8.k9.k25")
    assert result is None


def test_nested_json_value():
    result = dictor(NESTED, search="level")
    assert result == ['L10', 'L6', 'L5', 'L2', 'L1']

    result = dictor(NESTED, "k1.k2.k3.k4.k5.k6.k7.k8.k9.k10")
    assert isinstance(result, dict)


from __future__ import print_function
import json


def _pretty(data, pretty):
    """return output in pretty print"""
    if pretty:
        return json.dumps(data, indent=4, sort_keys=True)
    else:
        return data


def _findval(data, path, pathsep, ignorecase, default):
    """cycle through Dict and find the key's value"""
    val = default
    if r"\." in path:
        path = path.replace(r"\.", "__dictor__")

    for key in path.split(pathsep):
        if "__dictor__" in key:
            key = key.replace("__dictor__", ".")
        if isinstance(data, (list, tuple)):
            try:
                val = data[int(key)]
            except (UnboundLocalError, IndexError, ValueError):
                val = default
        else:
            if ignorecase:
                for datakey in data.keys():
                    if datakey.lower() == key.lower():
                        key = datakey
                        break
            try:
                if data and key in data:
                    val = data[key]
                else:
                    val = default
                    break
            except TypeError:
                val = default
        data = val
    return val

def _search(data, search, default):
    """search for specific keys, return list of values"""

    search_ret = []

    def recursive_search(searchdata, search):
        try:
            for k, v in searchdata.items():
                if k == search:
                    if default and v is None:
                        yield default
                    else:
                        yield v
                elif isinstance(v, dict):
                    for val in recursive_search(v, search):
                        yield val
                elif isinstance(v, list):
                    for list_item in v:
                        if isinstance(list_item, dict):
                            for val in recursive_search(list_item, search):
                                yield val
        except (KeyError, ValueError, IndexError, TypeError, AttributeError):
            pass

    if isinstance(data, (list, tuple)):
        for d in data:
            for v in recursive_search(d, search):
                search_ret.append(v)
    else:
        for v in recursive_search(data, search):
            search_ret.append(v)
    
    if search_ret:
        val = search_ret
    else:
        val = default
    return val

def dictor(
    data,
    path=None,
    default=None,
    checknone=False,
    ignorecase=False,
    pathsep=".",
    search=None,
    pretty=False,
    rtype=None,
):
    """Get a value or a list of values from a dictionary key using a path

    Args:
        data (dict | list): Input dictionary to be searched in.
        path (str, optional): Dictionary key search path (pathsep separated).
            Defaults to None.
        default (Any, optional): Default value to return if the key is not found.
            Defaults to None.
        checknone (bool, optional): If set, an exception is thrown if the value
            is None. Defaults to False.
        ignorecase (bool, optional): If set, upper/lower-case keys are treated
            the same. Defaults to False.
        pathsep (str, optional): Path separator for path parameter. Defaults to ".".
        search (Any, optional): Search for specific keys and output a list of values.
            Defaults to None.
        pretty (bool, optional): Pretty prints the result. Defaults to False.
        rtype (str, optional): Specifies in what type the value should be returned, either 'str' or 'int'.

    Raises:
        ValueError: Raises if checknone is set.

    Returns:
        Any: Returns one value or a list of values for the specified key.

    Examples:
        > dictor(data, "employees.John Doe.first_name")
        parse key index and fallback on default value if None,
        > dictor(data, "employees.5.first_name", "No employee found")
        pass a parameter
        > dictor(data, "company.{}.address".format(my_company))
        if using Python 3, can use F-strings to pass parameter
        > param = 'MyCompany'
        > dictor(data, f"company.{param}.address")
        lookup a 3rd element of List, on second key, lookup index=5
        > dictor(data, "3.first.second.5")
        lookup a nested list of lists
        > dictor(data, "0.first.1.2.second.third.0.2"
        check if return value is None, if it is, raise an error
        > dictor(data, "some.key.value", checknone=True)
        > ValueError: value not found for search path: "some.key.value"
        ignore letter casing when searching
        > dictor(data, "employees.Fred Flintstone", ignorecase=True)
        search data for specific keys (will return a list of all keys it finds)
        > dictor(data, "employees", search="name")
        pretty-print output
        > dictor(data, "employees", pretty=True)
        return a value in a specific type format
        > dictor(data, "employee_id", rtype="int")
    """
    if search is None and path is None:
        return _pretty(data, pretty)
    elif search and not path:
        val = _search(data, search, default)
    elif path:
        val = _findval(data, path, pathsep, ignorecase, default)
        if search and val and val != default:
            val = _search(val, search, default)
    else:
        val = default

    if checknone:
        if val is None or val == default:
            raise ValueError('value not found for search path: "%s"' % path)

    # return specific type
    if rtype and (type(val) is str or type(val) is int):
        try:
            if rtype == "int":
                val = int(val)
            if rtype == "str":
                val = str(val)
        except ValueError:
            pass
    return _pretty(val, pretty)


from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="dictor",
    version="0.1.12",
    description="an elegant dictionary and JSON handler",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/perfecto25/dictor",
    author="mike.reider",
    author_email="mike.reider@gmail.com",
    license="MIT",
    packages=["dictor"],
    zip_safe=True,
)
